import random
from logging import getLogger

from PIL import Image
import os
import torch
from PIL import ImageFilter, ImageOps
import numpy as np
import torchvision.transforms as transforms

logger = getLogger()


class ImageNetDataset(torch.utils.data.Dataset):
    def __init__(self, root, split, transform, size_dataset=-1):
        self.split = split
        self.root = root

        self.samples = []
        path_to_file = os.path.join(self.root, f"{self.split}.csv")
        with open(path_to_file, "r") as f:
            for clip_idx, path_label in enumerate(f.read().splitlines()):
                path, label = path_label.split(" ")
                item = os.path.join(self.root, path), int(label)
                self.samples.append(item)

        if size_dataset >= 0:
            self.samples = self.samples[:size_dataset]

        assert (len(self.samples) > 0), "Failed to load ImageNet split {} from {}".format(
            self._split, path_to_file)

        self.trans = transform

    def __getitem__(self, index):
        path, label = self.samples[index]

        img = Image.open(path).convert('RGB')
        if self.trans is not None:
            img = self.trans(img)
        return img, label

    def __len__(self):
        return len(self.samples)


class MultiCropDataset(ImageNetDataset):
    def __init__(
        self,
        data_path,
        split,
        size_crops,
        nmb_crops,
        min_scale_crops,
        max_scale_crops,
        size_dataset=-1,
        return_index=False,
        byolAug=False,
    ):
        super(MultiCropDataset, self).__init__(data_path, split, None)
        assert len(size_crops) == len(nmb_crops)
        assert len(min_scale_crops) == len(nmb_crops)
        assert len(max_scale_crops) == len(nmb_crops)
        if size_dataset >= 0:
            self.samples = self.samples[:size_dataset]
        self.return_index = return_index

        color_transform = [get_color_distortion(s=0.5 if byolAug else 1.0), PILRandomGaussianBlurWithSolarize(p=1.0 if byolAug else 0.5)]
        mean = [0.485, 0.456, 0.406]
        std = [0.228, 0.224, 0.225]
        trans = []
        for i in range(len(size_crops)):
            randomresizedcrop = transforms.RandomResizedCrop(
                size_crops[i],
                scale=(min_scale_crops[i], max_scale_crops[i]),
            )
            trans.extend([transforms.Compose([
                randomresizedcrop,
                transforms.RandomHorizontalFlip(p=0.5),
                transforms.Compose(color_transform),
                transforms.ToTensor(),
                transforms.Normalize(mean=mean, std=std)])
            ] * nmb_crops[i])
        self.trans = trans

        self.trans_k = None
        if byolAug:
            color_transform_k = [get_color_distortion(s=0.5),
                                 PILRandomGaussianBlurWithSolarize(p=0.1, p_solarize=0.2)]
            trans_k = []
            for i in range(len(size_crops)):
                randomresizedcrop = transforms.RandomResizedCrop(
                    size_crops[i],
                    scale=(min_scale_crops[i], max_scale_crops[i]),
                )
                trans_k.extend([transforms.Compose([
                    randomresizedcrop,
                    transforms.RandomHorizontalFlip(p=0.5),
                    transforms.Compose(color_transform_k),
                    transforms.ToTensor(),
                    transforms.Normalize(mean=mean, std=std)])
                ] * nmb_crops[i])
            self.trans_k = trans_k

    def __getitem__(self, index):
        path, _ = self.samples[index]
        image = Image.open(path).convert('RGB')
        multi_crops = list(map(lambda trans: trans(image), self.trans))
        if self.trans_k is not None:
            multi_crops_k = list(map(lambda trans: trans(image), self.trans_k))
        else:
            multi_crops_k = multi_crops
        if self.return_index:
            return index, multi_crops, multi_crops_k
        return multi_crops, multi_crops_k


class PILRandomGaussianBlurWithSolarize(object):
    """
    Apply Gaussian Blur to the PIL image. Take the radius and probability of
    application as the parameter.
    This transform was used in SimCLR - https://arxiv.org/abs/2002.05709
    """

    def __init__(self, p=0.5, p_solarize=0.0, radius_min=0.1, radius_max=2.):
        self.prob = p
        self.prob_solarize = p_solarize
        self.radius_min = radius_min
        self.radius_max = radius_max

    def __call__(self, img):
        do_it = np.random.rand() <= self.prob
        if do_it:
            img = img.filter(
            ImageFilter.GaussianBlur(
                radius=random.uniform(self.radius_min, self.radius_max)
            ))
        do_solarize = np.random.rand() <= self.prob_solarize
        if do_solarize:
            img = ImageOps.solarize(img)
        return img


def get_color_distortion(s=1.0):
    # s is the strength of color distortion.
    color_jitter = transforms.ColorJitter(0.8*s, 0.8*s, 0.4*s, 0.2*s)
    rnd_color_jitter = transforms.RandomApply([color_jitter], p=0.8)
    rnd_gray = transforms.RandomGrayscale(p=0.2)
    color_distort = transforms.Compose([rnd_color_jitter, rnd_gray])
    return color_distort
